// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright: 2009
//
//   Ulm University
// 
// Creator: Hendrik Lensch, Holger Dammertz
// Email:   hendrik.lensch@uni-ulm.de, holger.dammertz@
// ==========================================================================
// $Log$
// ==========================================================================

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

#include "PPM.hh"

using namespace std;
using namespace ppm; 

#define MAX_BLOCKS 512
#define MAX_THREADS 128 

//-------------------------------------------------------------------------------
  
// specify the gamma value to be applied
__device__ __constant__ float gpuGamma[1]; 

__device__ float applyGamma(const float& _src, const float _gamma) {
  return 255.0f * __powf(_src / 255.0f, _gamma); 
}
  

/* compute gamma correction on the float image _src of resolution dim,
   outputs the gamma corrected image should be stored in_dst[blockIdx.x *
   blockDim.x + threadIdx.x]. Each thread computes on pixel element.
*/ 
__global__ void gammaKernel(float *_dst, const float* _src, int _w) {
  int x = blockIdx.x * MAX_THREADS + threadIdx.x; 
  int y = blockIdx.y; 
  int pos = y * _w + x; 

  if (x < _w) {
     _dst[pos] = applyGamma(_src[pos], gpuGamma[0]); 
  }
}

//-------------------------------------------------------------------------------

float applyGammaHost(const float& _src, const float _gamma) {
  return 255.0f * powf(_src / 255.0f, _gamma); 
}

//-------------------------------------------------------------------------------
  
int main(int argc, char* argv[]) {
  CUT_DEVICE_INIT(1, argv); 
  int acount = 1; // parse command line 

  if (argc < 5) { 
	  printf("usage: %s <inImg> <gamma> <mode> <outImg>", argv[0]); 
    exit(1) ;
  }

  float* img;  

  int w,h;
  readPPM(argv[acount++], w, h, &img); 
 
  float gamma = atof(argv[acount++]); 
  // flag indicating weather the CPU or the GPU version should be executed
  bool gpuVersion = atoi(argv[acount++]); 

  int nPix = w*h; 

  float* gpuImg; 
  float* gpuResImg; 
  
  printf("  gamma: %f\n", gamma); 

  //-------------------------------------------------------------------------------
  if (gpuVersion) {
    printf("Executing the GPU Version\n");
	// copy the image to the device
	CUDA_SAFE_CALL(cudaMalloc((void**)&gpuImg, nPix* 3 * sizeof(float)));   
	CUDA_SAFE_CALL(cudaMalloc((void**)&gpuResImg, nPix* 3 * sizeof(float)));   
	CUDA_SAFE_CALL(cudaMemcpy(gpuImg, img, nPix * 3 * sizeof(float), cudaMemcpyHostToDevice));

	// copy gamma value to constant device memory
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(gpuGamma, &gamma, sizeof(float)));

	// calculate the block dimensions
	dim3 threadBlock(MAX_THREADS); 
	// select the number of blocks vertically (*3 because of RGB)
	dim3 blockGrid((w*3)/MAX_THREADS + 1, h, 1); 
	printf("bl/thr: %d  %d %d\n", blockGrid.x, blockGrid.y, threadBlock.x); 

	gammaKernel<<< blockGrid, threadBlock >>>(gpuResImg, gpuImg, w * 3); 
	CUT_CHECK_ERROR("gamma filter failed\n");
	CUDA_SAFE_CALL(cudaThreadSynchronize());	
	
	// download result
	CUDA_SAFE_CALL(cudaMemcpy(img, gpuResImg, nPix*3 * sizeof(float), cudaMemcpyDeviceToHost));
	CUT_CHECK_ERROR("copy failed\n");

	cudaFree(gpuResImg); 
	cudaFree(gpuImg); 
  } 
  //-------------------------------------------------------------------------------
  else {
    printf("Executing the CPU Version\n");
    for (int i = 0; i < nPix * 3; ++i) {
      img[i] = applyGammaHost(img[i], gamma); 
    }
  }

  writePPM(argv[acount++], w, h, (float*)img); 

  delete[] img; 

  printf("  done\n"); 
}

    


