// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright:
//
//   Max-Planck-Institut fuer Informatik
//   Stuhlsatzenhausweg 85, 66123 Saarbruecken
//   http://www.mpi-sb.mpg.de
// 
// Creator: lensch (Hendrik Lensch)
// Email:   lensch@mpi-sb.mpg.de
// ==========================================================================
// $Log$
// ==========================================================================

#ifndef EX1_TESTDOTPRODUCT_C
#define EX1_TESTDOTPRODUCT_C

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

using namespace std;
// using namespace base;
//using namespace image;

#define MAX_BLOCKS 256
#define MAX_THREADS 128

  
/* compute the dot product between a1 and a2. a1 and a2 are of size
dim. The result of each thread should be stored in _dst[blockIdx.x *
blockDim.x + threadIdx.x]. Each thread should accumulate the dot
product of a subset of elements. 
*/ 
__global__ void dotProdKernel(float *_dst, const float* _a1, const
			      float* _a2, int _dim) {

  const unsigned int outputIdx = blockIdx.x * MAX_THREADS + threadIdx.x;
  float reg = 0.f;
 
  for(unsigned int inputIdx = outputIdx; inputIdx < _dim; inputIdx += MAX_BLOCKS * MAX_THREADS)
  {
    reg += _a1[inputIdx] * _a2[inputIdx];
  }

  _dst[outputIdx] = reg;
}

  
/* This program sets up two large arrays of size dim and computes the
   dot product of both arrays. 

   The arrays are uploaed only once and the dot product is computed
   multiple times. While this does not make too much sense it
   demonstrated the possible speedup.  */ 

int main(int argc, char* argv[]) {

  // parse command line 
  int acount = 1; 

  if (argc < 3) { 
    printf("usage: testDotProduct <dim> <GPU-flag [0,1]>"); 
    exit(1) ;
  }

  // number of elements in both vectors 
  int dim = atoi(argv[acount++]); 

  // flag indicating weather the CPU or the GPU version should be executed
  bool gpuVersion = atoi( argv[acount++] ); 


  printf("dim: %d\n", dim); 

  float* cpuArray1 = new float[dim]; 
  float* cpuArray2 = new float[dim]; 

  // initialize the two arrays 
  for (int i = 0; i < dim; ++i) {
    
#ifdef RTEST 
    cpuArray1[i] = drand48(); 
    cpuArray2[i] = drand48();
#else 
     cpuArray1[i] = 2;
     cpuArray2[i] = i % 10; 
#endif 
  }

  // now the gpu stuff 
  float* gpuArray1; 
  float* gpuArray2; 
  float* gpuResult; 

  float* h; 
  // setup cuda environment
  if (gpuVersion) {
    // setup cuda environment
    //    CUT_DEVICE_INIT();	

    // allocate two gpuArray 1 and gpuArray 2 and gpuResult array on GPU
    CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuArray1, dim * sizeof(float) ) );   
    CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuArray2, dim * sizeof(float) ) );      
    CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResult, MAX_BLOCKS * MAX_THREADS * sizeof(float) ) );   

    // copy the array once to the device
    
    CUDA_SAFE_CALL( cudaMemcpy( gpuArray1, cpuArray1, dim * sizeof(float),  cudaMemcpyHostToDevice) );
    
    CUDA_SAFE_CALL( cudaMemcpy( gpuArray2, cpuArray2, dim * sizeof(float),  cudaMemcpyHostToDevice) );

    // allocate an array to download the results of all threads 
    h = new float[MAX_BLOCKS*MAX_THREADS]; 
  }

  for (int iter = 0; iter < 100; ++iter) {

    double finalDotProduct = 0.; 

    if (!gpuVersion) {
      printf( "cpu: "); 
      for (int i = 0; i < dim; ++i) {
	finalDotProduct +=  cpuArray1[i] * cpuArray2[i]; 

      }

    } 
    else {

      // Cuda version here
      printf( "gpu: "); 
      
      // copy the arrays
//       CUDA_SAFE_CALL( cudaMemcpy( gpuArray1, cpuArray1, dim * sizeof(float),  cudaMemcpyHostToDevice) );

//      CUDA_SAFE_CALL( cudaMemcpy( gpuArray2, cpuArray2, dim * sizeof(float),  cudaMemcpyHostToDevice) );

      // a simplistic way of splitting the problem into threads
      dim3 blockGrid( MAX_BLOCKS ); 
      dim3 threadBlock(MAX_THREADS); 
      
      // call the kernel 
      dotProdKernel<<< blockGrid, threadBlock>>>(gpuResult, gpuArray1, gpuArray2,dim); 

      // check that everything went soothly
      CUT_CHECK_ERROR("dot product execution failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );	

      // download and combine the results of multiple threads  
      CUDA_SAFE_CALL( cudaMemcpy( h, gpuResult, MAX_BLOCKS * MAX_THREADS * sizeof(float),  cudaMemcpyDeviceToHost) );
      
      finalDotProduct = 0.; 
      for (int i = 0; i < MAX_BLOCKS * MAX_THREADS; ++i) {
	finalDotProduct += h[i]; 

      }
    }

    printf( "%f\n", finalDotProduct); 
       
  }
  
  
  if (gpuVersion) {

    // cleanup GPU memory
    
     CUDA_SAFE_CALL( cudaFree(gpuResult) ); 
     CUDA_SAFE_CALL( cudaFree(gpuArray2) ); 
     CUDA_SAFE_CALL( cudaFree(gpuArray1) ); 

    delete[] h; 
  }

  delete[] cpuArray2; 
  delete[] cpuArray1; 

  printf("done\n"); 

}

    

      
      

      
   

      

      
      





#endif /* EX1_TESTDOTPRODUCT_C */







