// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright:
//
//   Max-Planck-Institut fuer Informatik
//   Stuhlsatzenhausweg 85, 66123 Saarbruecken
//   http://www.mpi-sb.mpg.de
// 
// Creator: lensch (Hendrik Lensch)
// Email:   lensch@mpi-sb.mpg.de
// ==========================================================================
// $Log$
// ==========================================================================

#ifndef EX1_TESTGAMMA_C
#define EX1_TESTGAMMA_C

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

#include "PPM.hh"

using namespace std;
using namespace ppm; 

#define THREADS 16

// global texture reference 
texture< float4, 2, cudaReadModeElementType> texImg; 

 
/* Compute the correlation between two images. _img1 is read from global memory, _img2 is loaded as a texture. 
   Each thread has a specific offset. It then computes the dot product of the shifted texture image with _img1. 
   The result image at location  (x,y) has the offset (offsX, offsY) 
              
   dst(x,y) = sum_(i,j)  _img1(i,j) * _img2(i+offsX, i+offsY), 
   
   where sum_(i,j) specifies summing over all pixels in the image. 
           
   Make sure to load the necessary data of _img1 into shared memory, i.e. the entire image block-by-block. This way it 
   can be shared between all the threads of one block.

   The second image is not loaded from shared memory but access via tex2D.  Accessing the texImg will be cached automatically. 
*/ 
__global__ void correlationKernel(float3 *_dst, const float3* _img1, int _w, int _h) {

  // get the shared memory
  extern __shared__ float3 shm[]; 

  // compute the position within the image 
  int x = blockIdx.x * blockDim.x + threadIdx.x; 
  int y = blockIdx.y * blockDim.y + threadIdx.y; 

  // compute the offset of this thread
  int offsX = x - _w/2; 
  int offsY = y - _h/2; 

 
  int threadShmPos = threadIdx.x + threadIdx.y * blockDim.x;  

  float3 res; 
  res.x = 0.; 
  res.y = 0.; 
  res.z = 0.; 

  // loop over the image block by block  
  for (int v = 0; v < gridDim.y; ++v) {
    for (int u = 0; u < gridDim.x; ++u) {

      int baseX = u * blockDim.x; 
      int baseY = v * blockDim.y; 
      
      const float3* base = _img1 + (baseX + baseY * _w); 

      // load this block into shared shm
      shm[threadShmPos] = base[threadIdx.x + threadIdx.y * _w]; 

      // make sure that the shared memory has been written by all threads
      __syncthreads(); 
      
      // for each pixel in the block 
      for (int j = 0; j < THREADS; ++j) 
      {
	for (int i = 0; i < THREADS; ++i) {

	  int bj =  baseY + j + offsY;
	  int bi =  baseX + i + offsX;
	  int shmPos = i + j * THREADS; 

	  // get the pixel that is shifted by the offset from the
	  float4 r = tex2D( texImg, bi / (_w-1.), bj / (_h-1.) ); 

	  // multiply with shm pixel and add to result
 	  res.x += shm[shmPos].x * r.x; 
 	  res.y += shm[shmPos].y * r.y; 
 	  res.z += shm[shmPos].z * r.z; 

	}
      }

    }
  }


  // return only the positive values and scale to fit the output. 
  _dst[ x + y * _w ].x = max(0.,res.x) * 0.05; 
  _dst[ x + y * _w ].y = max(0.,res.y) * 0.05; 
  _dst[ x + y * _w ].z = max(0.,res.z) * 0.05; 

}




/** this function simply subtracts the average value from an image and scales it to the range of [0,1] from [0,255]; 
 */ 
void normalize( float* _img, int _w, int _h) {

  // first, compute the average
  float avg[3] = {0.,0.,0.}; 

  float* ii = _img; 
  
  for (int y = 0; y < _h; ++y) {
    for (int x = 0; x < _w; ++x) {

      avg[0] += *(ii++); 
      avg[1] += *(ii++); 
      avg[2] += *(ii++); 
    }
  }

  avg[0] /= (float)_w*_h; 
  avg[1] /= (float)_w*_h; 
  avg[2] /= (float)_w*_h; 

  // now subtract average 

  ii = _img; 
  for (int y = 0; y < _h; ++y) {
    for (int x = 0; x < _w; ++x) {

      *(ii++) -= avg[0]; 
      *(ii++) -= avg[1]; 
      *(ii++) -= avg[2];     
    }
  }

  ii = _img; 
  for (int y = 0; y < _h; ++y) {
    for (int x = 0; x < _w * 3; ++x) {
      *(ii++) /= 255.;
    }
  }      


}
  
  

      
    

  
/* This program computes the cross-correlation between two images of
   the same size.  The size of the test images is nicely chosen to be
   multiples of 16 in each dimension, so don't worry about image
   boundaries! 

   All computations have to be done in the image domain (there exist
   faster ways to calculate the cross-correlation in Fourier
   domain).
*/ 

int main(int argc, char* argv[]) {

  // parse command line 
  int acount = 1; 

  if (argc < 4) { 
    printf("usage: testGamma <inImg> <inImg2> <outImg>"); 
    exit(1) ;
  }

  float* img1;  
  float* img2;  

  int w,h;
  readPPM( argv[acount++], w, h, &img1 ); 
  readPPM( argv[acount++], w, h, &img2 ); 

  // let's compute normalized cross products
  // so we normalize both images, i.e. subtract the mean value and scale by 1./255.
  normalize( img1, w,h); 
  normalize( img2, w,h); 

  int nPix = w*h; 

  float* gpuImg1; 
  float* gpuResImg; 
 
  // copy the image to the device
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuImg1, nPix* 3 *  sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResImg, nPix* 3 * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMemcpy( gpuImg1, img1, nPix*3 * sizeof(float),
  cudaMemcpyHostToDevice) );


  // furthermore, the second image is accessed as a texture 

  // set the texture mode to normalized 
  texImg.normalized = 1; 

  // and allow texture coordinates to wrap at boundaries 
  texImg.addressMode[0] = cudaAddressModeWrap; 
  texImg.addressMode[1] = cudaAddressModeWrap; 


  cudaArray* gpuTex; 

  cudaChannelFormatDesc floatTex = cudaCreateChannelDesc<float4>();
  CUDA_SAFE_CALL( cudaMallocArray( &gpuTex, &floatTex, w, h) ); 

  // pad to float4 for faster access 
  float* img4 = new float[w*h*4]; 

  for (int i = 0; i < w*h; ++i) {

    img4[4*i] = img2[3*i]; 
    img4[4*i+1] = img2[3*i+1]; 
    img4[4*i+2] = img2[3*i+2]; 
    img4[4*i+3] = 0.; 
  }

  // upload to array
  CUDA_SAFE_CALL( cudaMemcpyToArray(gpuTex, 0, 0, img4, w*h*4*sizeof(float), cudaMemcpyHostToDevice) );
  
  // bind as texture
  CUDA_SAFE_CALL(cudaBindTextureToArray( texImg, gpuTex, floatTex) );

  
  // calculate the block dimensions
  dim3 threadBlock( THREADS, THREADS ); 
  dim3 blockGrid( w / THREADS , h / THREADS,  1); 

  printf( "blockDim: %d  %d \n", threadBlock.x, threadBlock.y ); 
  printf( "gridDim: %d  %d \n", blockGrid.x, blockGrid.y ); 

  int shmSize = threadBlock.x * threadBlock.y * 4 * sizeof(float); 
  correlationKernel<<< blockGrid, threadBlock, shmSize >>>( (float3*)gpuResImg, (float3*)gpuImg1, w, h ); 
  
  CUT_CHECK_ERROR("correlation filter failed\n");
  CUDA_SAFE_CALL( cudaThreadSynchronize() );	
  
  // download result
  CUDA_SAFE_CALL( cudaMemcpy( img1, gpuResImg, nPix*3 * sizeof(float),  cudaMemcpyDeviceToHost) );
  CUT_CHECK_ERROR("copy failed\n");
   

  CUDA_SAFE_CALL(cudaUnbindTexture(texImg));
  cudaFreeArray( gpuTex ); 

  cudaFree( gpuResImg ); 
  cudaFree( gpuImg1 ); 

  writePPM( argv[acount++], w, h, (float*)img1 ); 

  delete[] img1; 
  delete[] img2; 

  delete[] img4; 

  printf("done\n"); 

}

    

      
      

      
   

      

      
      





#endif /* EX1_TESTGAMMA_C */







