// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright:
//
//   Max-Planck-Institut fuer Informatik
//   Stuhlsatzenhausweg 85, 66123 Saarbruecken
//   http://www.mpi-sb.mpg.de
// 
// Creator: lensch (Hendrik Lensch)
// Email:   lensch@mpi-sb.mpg.de
// ==========================================================================
// $Log$
// ==========================================================================

#ifndef EX1_TESTDOTPRODUCT_C
#define EX1_TESTDOTPRODUCT_C

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

using namespace std;
// using namespace base;
//using namespace image;

#define MAX_BLOCKS 256
#define MAX_THREADS 128

  
__global__ void dotProdKernel(float *_dst, const float* _a1, const
			      float* _a2, int _dim) {
  
  // calculate how many elements each thread needs to calculate 
  const unsigned int iter = _dim / (blockDim.x * gridDim.x); 
  int pos = blockIdx.x*MAX_THREADS + threadIdx.x; 
 
  // clear the output 
  _dst[ blockIdx.x*MAX_THREADS + threadIdx.x] = 0; 

  for (int i = 0; i < iter; ++i) {

    _dst[ blockIdx.x*MAX_THREADS + threadIdx.x] += _a1[pos] * _a2[pos];
    pos += blockDim.x * gridDim.x;  
  }
 
  // for the last iteration, check if the elements are still available
  if (pos < _dim) {
    _dst[ blockIdx.x*MAX_THREADS + threadIdx.x] += _a1[pos] * _a2[pos];
  }
   
}

  
/* This program sets up two large arrays of size dim and computes the
   dot product of both arrays. 

   The kernel from the last exercise can be reused. 
   In contrast to the last exercise the second array will be uploaded in each iteration. 

   In this exercise you should implement the upload of the second array in three different ways: 

   1. using simple memcopy (as in exercise 1)
   2. using memcopy from non-pageable memory (should be faster)
   3. using asynchronous memory copy with two streams (you can compute the result of each stream into the h[0], h[1]. 

*/ 
   

int main(int argc, char* argv[]) {

  // parse command line 
  int acount = 1; 

  if (argc < 3) { 
    printf("usage: testDotProductStreams <dim> <copy mode [0,1,2]>"); 
    exit(1) ;
  }

  // number of elements in both vectors 
  int dim = atoi(argv[acount++]); 

  // mode of the memory upload
  int mode = atoi( argv[acount++] ); 


  printf("dim: %d\n", dim); 

  float* cpuArray1 = new float[dim]; 
  float* cpuArray2 = new float[dim]; 

  // extra array that points to non-pagelocked array for faster transfer
  float* cpuArray2Pagelocked; 

  // create the pagelocked array of dimension dim
  cudaMallocHost( (void**)&cpuArray2Pagelocked, dim * sizeof(float)); 

  // initialize the two arrays 
  for (int i = 0; i < dim; ++i) {
    
#ifdef RTEST 
    cpuArray1[i] = drand48(); 
    cpuArray2[i] = drand48();
#else 
     cpuArray1[i] = 1.0;
     cpuArray2[i] = i; 
     cpuArray2Pagelocked[i] = i; 
#endif 
  }

  // now the gpu stuff 
  float* gpuArray1; 
  float* gpuArray2; 

  // you might need two result arrays for the kernels of the two streams
  float* gpuResult[2]; 
  float* h[2]; 



  // setup cuda environment

  // CUT_DEVICE_INIT();	

  // create two streams for the last mode
  cudaStream_t mystream[2];
  
  for (int i = 0; i < 2; ++i) {
      cudaStreamCreate(&(mystream[i])); 
  }
  
  // allocate two gpuArray 1  and gpuArray 2 and gpuResult array on GPU
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuArray1, dim * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuArray2, dim * sizeof(float) ) );      
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResult[0], MAX_BLOCKS * MAX_THREADS * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResult[1], MAX_BLOCKS * MAX_THREADS * sizeof(float) ) );   
  
  // copy array 1 once to the device
  CUDA_SAFE_CALL( cudaMemcpy( gpuArray1, cpuArray1, dim * sizeof(float),  cudaMemcpyHostToDevice) );
  
  
  // allocate an array to download the results of all threads 
  cudaMallocHost( (void**)&h[0], MAX_THREADS*MAX_BLOCKS * sizeof(float)); 
  cudaMallocHost( (void**)&h[1], MAX_THREADS*MAX_BLOCKS * sizeof(float)); 


  for (int iter = 0; iter < 100; ++iter) {

    double finalDotProduct = 0.; 

    // a simplistic way of splitting the problem into threads
    dim3 blockGrid( MAX_BLOCKS ); 
    dim3 threadBlock(MAX_THREADS); 


    switch (mode) {

    case 0: 
      // copy a simple array 
      printf( "simple memcpy: "); 
      
      // update Array 2
      CUDA_SAFE_CALL( cudaMemcpy( gpuArray2, cpuArray2, dim * sizeof(float),  cudaMemcpyHostToDevice) );

      
      // call the kernel 
      dotProdKernel<<< blockGrid, threadBlock>>>(gpuResult[0], gpuArray1, gpuArray2,dim); 

      // check that everything went smooth
      CUT_CHECK_ERROR("dot product execution failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );	

      // download and combine the results of multiple threads  
      CUDA_SAFE_CALL( cudaMemcpy( h[0], gpuResult[0], MAX_BLOCKS * MAX_THREADS * sizeof(float),  cudaMemcpyDeviceToHost) );
      
      finalDotProduct = 0.; 
      for (int i = 0; i < MAX_BLOCKS * MAX_THREADS; ++i) {
	finalDotProduct += h[0][i]; 
      }

      break; 
 
    case 1: 

      // copy Array2 from pagelocked memory
      printf( "pagelocked memory: "); 
      
      // copy the arrays
      //CUDA_SAFE_CALL( cudaMemcpy( gpuArray1, cpuArray1, dim * sizeof(float),  cudaMemcpyHostToDevice) );

      CUDA_SAFE_CALL( cudaMemcpy( gpuArray2, cpuArray2Pagelocked, dim * sizeof(float),  cudaMemcpyHostToDevice) );

      
      // call the kernel 
      dotProdKernel<<< blockGrid, threadBlock>>>(gpuResult[0], gpuArray1, gpuArray2,dim); 

      // check that everything went soothly
      CUT_CHECK_ERROR("dot product execution failed\n");
      CUDA_SAFE_CALL( cudaThreadSynchronize() );	

      // download and combine the results of multiple threads  
      CUDA_SAFE_CALL( cudaMemcpy( h[0], gpuResult[0], MAX_BLOCKS * MAX_THREADS * sizeof(float),  cudaMemcpyDeviceToHost) );
      
      finalDotProduct = 0.; 
      for (int i = 0; i < MAX_BLOCKS * MAX_THREADS; ++i) {
	finalDotProduct += h[0][i]; 
      }

      break; 

    case 2: 

      // use two streams with interleaved processing 
      // use asynchronous up and download
      printf( "2 streams: "); 
      
      // copy the two halves of Array2 asynchronously 
      // by linking the upload to the streams
      // asynchronous copy requires page locked memory

      CUDA_SAFE_CALL( cudaMemcpyAsync( gpuArray2, cpuArray2Pagelocked, dim / 2 * sizeof(float),  cudaMemcpyHostToDevice, mystream[0]) );

      CUDA_SAFE_CALL( cudaMemcpyAsync( gpuArray2 + dim/2, cpuArray2Pagelocked + dim/2, dim / 2 * sizeof(float),  cudaMemcpyHostToDevice, mystream[1]) );
      
      // call the kernel twice, once for each stream 

      //      CUDA_SAFE_CALL( cudaStreamSynchronize(mystream[0]) );	
      dotProdKernel<<< blockGrid, threadBlock, 0, mystream[0]>>>(gpuResult[0], gpuArray1, gpuArray2,dim/2); 
      //      CUDA_SAFE_CALL( cudaStreamSynchronize(mystream[1]) );	
      dotProdKernel<<< blockGrid, threadBlock, 0, mystream[1]>>>(gpuResult[1], gpuArray1+dim/2, gpuArray2 + dim/2 ,dim/2); 

      // check that everything went soothly
      CUT_CHECK_ERROR("dot product execution failed\n");

      //      CUDA_SAFE_CALL( cudaThreadSynchronize() );	

      // download and combine the results of the two threads       
      CUDA_SAFE_CALL( cudaMemcpyAsync( h[0], gpuResult[0], MAX_BLOCKS
      * MAX_THREADS * sizeof(float),  cudaMemcpyDeviceToHost, mystream[0]) );
      CUDA_SAFE_CALL( cudaMemcpyAsync( h[1], gpuResult[1], MAX_BLOCKS
      * MAX_THREADS * sizeof(float),  cudaMemcpyDeviceToHost, mystream[1]) );

      //      CUDA_SAFE_CALL( cudaThreadSynchronize() );	

      finalDotProduct = 0.; 

      // compute the final product

      CUDA_SAFE_CALL( cudaStreamSynchronize(mystream[0]) );	
      for (int i = 0; i < MAX_BLOCKS * MAX_THREADS; ++i) {
	finalDotProduct += h[0][i]; 
      }

      CUDA_SAFE_CALL( cudaStreamSynchronize(mystream[1]) );	
      for (int i = 0; i < MAX_BLOCKS * MAX_THREADS; ++i) {
	finalDotProduct += h[1][i]; 
      }


      break;


    } // end switch
      

    printf( "%f\n", finalDotProduct); 
       
  }
  
  
  for (int i = 0; i < 2; ++i) {
    cudaStreamDestroy( mystream[i] ); 
  }
  
  
  // cleanup GPU memory
  
  CUDA_SAFE_CALL( cudaFree(gpuResult[1]) ); 
  CUDA_SAFE_CALL( cudaFree(gpuResult[0]) ); 
  CUDA_SAFE_CALL( cudaFree(gpuArray2) ); 
  CUDA_SAFE_CALL( cudaFree(gpuArray1) ); 

  // free page locked memory
  cudaFreeHost( cpuArray2Pagelocked ); 
  cudaFreeHost( h[1] );   
  cudaFreeHost( h[0] ); 


  delete[] cpuArray2; 
  delete[] cpuArray1; 

  printf("done\n"); 

}

    

      
      

      
   

      

      
      





#endif /* EX1_TESTDOTPRODUCT_C */







