// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright:
//
//   Max-Planck-Institut fuer Informatik
//   Stuhlsatzenhausweg 85, 66123 Saarbruecken
//   http://www.mpi-sb.mpg.de
// 
// Creator: lensch (Hendrik Lensch)
// Email:   lensch@mpi-sb.mpg.de
// ==========================================================================
// $Log$
// ==========================================================================

#ifndef EX1_TESTGAMMA_C
#define EX1_TESTGAMMA_C

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>
#include <vector>
#include <iostream>
#include <string>

#include "PPM.hh"

using namespace std;
using namespace ppm; 

__device__ __constant__ float3 gpuClusterCol[2048];

#define THREADS 256
#define LOG_IMG_SIZE 8
#define IMG_SIZE 256
#define WINDOW 6

// global texture reference 
texture< float4, 2, cudaReadModeElementType> texImg; 


 
/* The function measures for every pixel the distance to all
   clusters, and determines the clusterID of the nearest cluster
   center. It then colors the pixel in the cluster's color. 

   The cluster centers are given as an array of linear indices into
   the vector image, i.e.    _clusterInfo[0] = (x_0 + y_0 * _w). 

*/ 
__global__ void voronoiKernel(float3 *_dst, int _w, int _h, int _nClusters, const int* _clusterInfo) {

  // get the shared memory
  extern __shared__ int shm[]; 

  int nIter = _nClusters / THREADS +1 ; 
  // load cluster data
  for (int i = 0; i < nIter; ++i) {
    int pos = i * THREADS + threadIdx.x; 
    if (pos < _nClusters) {
      shm[pos] = _clusterInfo[pos]; 
    }
  }
  
  __syncthreads(); 

  // compute the position within the image 
  float x = blockIdx.x * blockDim.x + threadIdx.x; 
  float y = blockIdx.y; 

  int pos = x + y * _w; 


  // determine which is the closest cluster
  float minDist = 1000000.; 
  int minIdx = 0; 
  for (int i = 0; i < _nClusters; ++i) {

    float yy = shm[i] >> LOG_IMG_SIZE; 
    float xx = shm[i] % IMG_SIZE; 

    float dist = (x-xx)*(x-xx) + (y-yy)*(y-yy); 
    if (dist < minDist) {
      minDist = dist; 
      minIdx = i; 
    }
  }
  
  _dst[pos].x = gpuClusterCol[minIdx].x; 
  _dst[pos].y = gpuClusterCol[minIdx].y; 
  _dst[pos].z = gpuClusterCol[minIdx].z; 

  
  // mark the center of each cluster
  if (minDist <= 2.) {
    _dst[pos].x = 255;
    _dst[pos].y = 0.;
    _dst[pos].z = 0.; 
  }
      

}


__device__ float luminance( const float4& _col ) {
  return 0.299 *_col.x + 0.587*_col.y + 0.114*_col.z; 
}



/** stores a 1 in _dst if the pixel's luminance is a maximum in the
    WINDOW x WINDOW neighborhood 
*/ 
__global__ void featureKernel(float *_dst, int _w, int _h) {

  
  // compute the position within the image 
  int x = blockIdx.x * blockDim.x + threadIdx.x; 
  int y = blockIdx.y; 

  float lum = luminance( tex2D(texImg, x,y) ); 

  bool maximum = false; 

  if (lum > 20) {
    maximum = true; 
    for (int v = y -WINDOW; v < y +WINDOW; ++v) {    
      for (int u = x -WINDOW; u < x +WINDOW; ++u) {
	
	if (lum < luminance( tex2D(texImg, u,v))) {
	  maximum = false; 
	}
	
      }
    }
  }

  if (maximum) {
    _dst[x + y * _w] = 1.;
  } else {
    _dst[x + y * _w] = 0.;
  }
    
}

    

  
/* This program detects the local maxima in an image, writes their
   location into a vector and then computes the Voronoi diagram of the
   image given the detected local maxima as cluster centers. 

   A Voronoi diagram simply colors every pixel with the color of the
   nearest cluster center. */ 

int main(int argc, char* argv[]) {

  // parse command line 
  int acount = 1; 

  if (argc < 3) { 
    printf("usage: testPrefix <inImg> <inImg2> <mode>"); 
    exit(1) ;
  }

  float* img1;  

  int w,h;
  readPPM( argv[acount++], w, h, &img1 ); 

  string outName( argv[acount++] ); 

  int mode = atoi(argv[acount++]); 

  int nPix = w*h; 

  // result image of the feature detector
  // a single channel image, contains 1 for the detected maxima, 0 otherwise
  float* gpuFeatureImg;    
  
  // final result image of the Voronoi diagram 
  float3* gpuVoronoiImg; 

  // compacted feature list (idx within the image)
  // this is where your algorithm needs to store the compacted array
  int* gpuFeatureList; 
 
  // copy the image to the device
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuFeatureImg, nPix *  sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuVoronoiImg, nPix* 3 * sizeof(float) ) );   

  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuFeatureList, 10000 * sizeof(int) ) );   

  // color map for the cluster
  float clusterCol[2048*3]; 

  float* ci=clusterCol; 
  for (int i = 0; i < 2048; ++i, ci+=3) {
    ci[0] = 32*i % 256;
    ci[1] = (10*i +128) % 256; 
    ci[2] = (40*i +255) % 256;
  }

  CUDA_SAFE_CALL( cudaMemcpyToSymbol(gpuClusterCol, clusterCol, 2048*3*sizeof(float)) );
  
  cudaArray* gpuTex; 
  cudaChannelFormatDesc floatTex = cudaCreateChannelDesc<float4>();
  CUDA_SAFE_CALL( cudaMallocArray( &gpuTex, &floatTex, w, h) ); 

  // pad to float4 for faster access 
  float* img4 = new float[w*h*4]; 

  for (int i = 0; i < w*h; ++i) {

    img4[4*i] = img1[3*i]; 
    img4[4*i+1] = img1[3*i+1]; 
    img4[4*i+2] = img1[3*i+2]; 
    img4[4*i+3] = 0.; 
  }

  // upload to array
  CUDA_SAFE_CALL( cudaMemcpyToArray(gpuTex, 0, 0, img4, w*h*4*sizeof(float), cudaMemcpyHostToDevice) );
  
  // bind as texture
  CUDA_SAFE_CALL(cudaBindTextureToArray( texImg, gpuTex, floatTex) );

  cout << "setup texture" << endl; cout.flush(); 
  
  // calculate the block dimensions
  dim3 threadBlock( THREADS ); 
  dim3 blockGrid( w / THREADS , h ,  1); 

  printf( "blockDim: %d  %d \n", threadBlock.x, threadBlock.y ); 
  printf( "gridDim: %d  %d \n", blockGrid.x, blockGrid.y ); 
  
  featureKernel<<< blockGrid, threadBlock >>>( (float*)gpuFeatureImg, w, h ); 
  
  CUT_CHECK_ERROR("feature detector failed\n");
  CUDA_SAFE_CALL( cudaThreadSynchronize() );	


  // variable to store the number of detected features = the number of clusters
  int nFeatures; 


  if (mode == 0) {

    ////////////////////////////////////////////////////////////
    // CPU compaction: 
    ////////////////////////////////////////////////////////////

    // download result
    CUDA_SAFE_CALL( cudaMemcpy( img1, gpuFeatureImg, nPix * sizeof(float),  cudaMemcpyDeviceToHost) );
    CUT_CHECK_ERROR("copy failed\n");

    std::vector<int> features; 
    
    float* ii = img1; 
    for (int i = 0; i < nPix; ++i, ++ii) {
      if (*ii > 0.) {
	features.push_back( i ); 
	
	int x = i % IMG_SIZE; 
	int y = i >> LOG_IMG_SIZE; 
	//      std::cout << x << " " << y << std::endl; 
      }
      
    }
    
    cout << "nFeatures:  " << features.size() << endl ;

#if 0 
    // if you want to look at the feature image
    for (int i = nPix-1; i >= 0; --i) {
      img1[3*i] = 255*img1[i]; 
      img1[3*i+1] = 0; 
      img1[3*i+2] = 0; 
    }

    writePPM( "feature.ppm", w, h, (float*)img1 );
#endif 


    nFeatures = features.size(); 
    // upload feature vector
    CUDA_SAFE_CALL( cudaMemcpy( gpuFeatureList, &(features[0]), nFeatures * sizeof(int),  cudaMemcpyHostToDevice ) );

  } else {

    ////////////////////////////////////////////////////////////
    // GPU compaction: 
    ////////////////////////////////////////////////////////////

    // implement the prefixSum algorithm 
    // 1. Each block is simply computing the prefix sum of each scanline and outputs the total sum of the block 
    // 2. The a single block computes the prefix sum of the block sums. This block will compute the sum of all features -> nFeatures; 
    // 3. Add the resulting number from step 2 to the prefix sum of each block.
    //!!!!! missing !!!!! 
    



  }





  // now compute the Voronoi Diagram around the detected features. 
  voronoiKernel<<< blockGrid, threadBlock, nFeatures*sizeof(int) >>>( gpuVoronoiImg, w, h, nFeatures, gpuFeatureList ); 
  
  CUT_CHECK_ERROR("voronoi kernel failed\n");
  CUDA_SAFE_CALL( cudaThreadSynchronize() );	
	  

  // download final voronoi image. 
  CUDA_SAFE_CALL( cudaMemcpy( img1, gpuVoronoiImg, nPix * 3 * sizeof(float),  cudaMemcpyDeviceToHost) );
    

  CUDA_SAFE_CALL(cudaUnbindTexture(texImg));
  cudaFreeArray( gpuTex ); 

  cudaFree( gpuFeatureList ); 
  cudaFree( gpuFeatureImg ); 
  cudaFree( gpuVoronoiImg ); 

  //  writePPM( argv[acount++], w, h, (float*)img1 );
  writePPM( outName.c_str(), w, h, (float*)img1 );
 

  delete[] img1; 

  delete[] img4; 

  printf("done\n"); 

}

    

      
      

      
   

      

      
      





#endif /* EX1_TESTGAMMA_C */







